""" PyTorch distributed helpers

Some of this lifted from Detectron2 with other fns added by myself. Some of the Detectron2 fns
were intended for use with GLOO PG. I am using NCCL here with default PG so not everything will work
as is -RW
"""
import functools
import logging
import numpy as np
import pickle
import torch
import torch.distributed as dist

_LOCAL_PROCESS_GROUP = None
"""
A torch process group which only includes processes that on the same machine as the current process.
This variable is set when processes are spawned by `launch()` in "engine/launch.py".
"""


def get_world_size() -> int:
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1
    return dist.get_world_size()


def get_rank() -> int:
    if not dist.is_available():
        return 0
    if not dist.is_initialized():
        return 0
    return dist.get_rank()


def get_local_rank() -> int:
    """
    Returns:
        The rank of the current process within the local (per-machine) process group.
    """
    if not dist.is_available():
        return 0
    if not dist.is_initialized():
        return 0
    assert _LOCAL_PROCESS_GROUP is not None
    return dist.get_rank(group=_LOCAL_PROCESS_GROUP)


def get_local_size() -> int:
    """
    Returns:
        The size of the per-machine process group,
        i.e. the number of processes per machine.
    """
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1
    return dist.get_world_size(group=_LOCAL_PROCESS_GROUP)


def is_main_process() -> bool:
    return get_rank() == 0


def synchronize():
    """
    Helper function to synchronize (barrier) among all processes when
    using distributed training
    """
    if not dist.is_available():
        return
    if not dist.is_initialized():
        return
    world_size = dist.get_world_size()
    if world_size == 1:
        return
    dist.barrier()


@functools.lru_cache()
def _get_global_gloo_group():
    """
    Return a process group based on gloo backend, containing all the ranks
    The result is cached.
    """
    if dist.get_backend() == "nccl":
        return dist.new_group(backend="gloo")
    else:
        return dist.group.WORLD


def _serialize_to_tensor(data, group):
    backend = dist.get_backend(group)
    assert backend in ["gloo", "nccl"]
    device = torch.device("cpu" if backend == "gloo" else "cuda")

    buffer = pickle.dumps(data)
    if len(buffer) > 1024 ** 3:
        logger = logging.getLogger(__name__)
        logger.warning(
            "Rank {} trying to all-gather {:.2f} GB of data on device {}".format(
                get_rank(), len(buffer) / (1024 ** 3), device
            )
        )
    storage = torch.ByteStorage.from_buffer(buffer)
    tensor = torch.ByteTensor(storage).to(device=device)
    return tensor


def _pad_to_largest_tensor(tensor, group):
    """
    Returns:
        list[int]: size of the tensor, on each rank
        Tensor: padded tensor that has the max size
    """
    world_size = dist.get_world_size(group=group)
    assert (
        world_size >= 1
    ), "comm.gather/all_gather must be called from ranks within the given group!"
    local_size = torch.tensor([tensor.numel()], dtype=torch.int64, device=tensor.device)
    size_list = [
        torch.zeros([1], dtype=torch.int64, device=tensor.device) for _ in range(world_size)
    ]
    dist.all_gather(size_list, local_size, group=group)
    size_list = [int(size.item()) for size in size_list]

    max_size = max(size_list)

    # we pad the tensor because torch all_gather does not support
    # gathering tensors of different shapes
    if local_size != max_size:
        padding = torch.zeros((max_size - local_size,), dtype=torch.uint8, device=tensor.device)
        tensor = torch.cat((tensor, padding), dim=0)
    return size_list, tensor


def all_gather(data, group=None):
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors).
    Args:
        data: any picklable object
        group: a torch process group. By default, will use a group which
            contains all ranks on gloo backend.
    Returns:
        list[data]: list of data gathered from each rank
    """
    if get_world_size() == 1:
        return [data]
    if group is None:
        group = _get_global_gloo_group()
    if dist.get_world_size(group) == 1:
        return [data]

    tensor = _serialize_to_tensor(data, group)

    size_list, tensor = _pad_to_largest_tensor(tensor, group)
    max_size = max(size_list)

    # receiving Tensor from all ranks
    tensor_list = [torch.empty((max_size,), dtype=torch.uint8, device=tensor.device) for _ in size_list]
    dist.all_gather(tensor_list, tensor, group=group)

    data_list = []
    for size, tensor in zip(size_list, tensor_list):
        buffer = tensor.cpu().numpy().tobytes()[:size]
        data_list.append(pickle.loads(buffer))

    return data_list


def gather(data, dst=0, group=None):
    """
    Run gather on arbitrary picklable data (not necessarily tensors).
    Args:
        data: any picklable object
        dst (int): destination rank
        group: a torch process group. By default, will use a group which
            contains all ranks on gloo backend.
    Returns:
        list[data]: on dst, a list of data gathered from each rank. Otherwise,
            an empty list.
    """
    if get_world_size() == 1:
        return [data]
    if group is None:
        group = _get_global_gloo_group()
    if dist.get_world_size(group=group) == 1:
        return [data]
    rank = dist.get_rank(group=group)

    tensor = _serialize_to_tensor(data, group)
    size_list, tensor = _pad_to_largest_tensor(tensor, group)

    # receiving Tensor from all ranks
    if rank == dst:
        max_size = max(size_list)
        tensor_list = [torch.empty((max_size,), dtype=torch.uint8, device=tensor.device) for _ in size_list]
        dist.gather(tensor, tensor_list, dst=dst, group=group)

        data_list = []
        for size, tensor in zip(size_list, tensor_list):
            buffer = tensor.cpu().numpy().tobytes()[:size]
            data_list.append(pickle.loads(buffer))
        return data_list
    else:
        dist.gather(tensor, [], dst=dst, group=group)
        return []


def shared_random_seed():
    """
    Returns:
        int: a random number that is the same across all workers.
            If workers need a shared RNG, they can use this shared seed to
            create one.
    All workers must call this function, otherwise it will deadlock.
    """
    ints = np.random.randint(2 ** 31)
    all_ints = all_gather(ints)
    return all_ints[0]


def reduce_dict(input_dict, average=True):
    """
    Reduce the values in the dictionary from all processes so that process with rank
    0 has the reduced results.
    Args:
        input_dict (dict): inputs to be reduced. All the values must be scalar CUDA Tensor.
        average (bool): whether to do average or sum
    Returns:
        a dict with the same keys as input_dict, after reduction.
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict
    with torch.no_grad():
        names = []
        values = []
        # sort the keys so that they are consistent across processes
        for k in sorted(input_dict.keys()):
            names.append(k)
            values.append(input_dict[k])
        values = torch.stack(values, dim=0)
        dist.reduce(values, dst=0)
        if dist.get_rank() == 0 and average:
            # only main process gets accumulated, so only divide by
            # world_size in this case
            values /= world_size
        reduced_dict = {k: v for k, v in zip(names, values)}
    return reduced_dict


def all_gather_container(container, group=None, cat_dim=0):
    group = group or dist.group.WORLD
    world_size = dist.get_world_size(group)

    def _do_gather(tensor):
        tensor_list = [torch.empty_like(tensor) for _ in range(world_size)]
        dist.all_gather(tensor_list, tensor, group=group)
        return torch.cat(tensor_list, dim=cat_dim)

    if isinstance(container, dict):
        gathered = dict()
        for k, v in container.items():
            v = _do_gather(v)
            gathered[k] = v
        return gathered
    elif isinstance(container, (list, tuple)):
        gathered = [_do_gather(v) for v in container]
        if isinstance(container, tuple):
            gathered = tuple(gathered)
        return gathered
    else:
        # if not a dict, list, tuple, expect a singular tensor
        assert isinstance(container, torch.Tensor)
        return _do_gather(container)


def gather_container(container, dst, group=None, cat_dim=0):
    group = group or dist.group.WORLD
    world_size = dist.get_world_size(group)
    this_rank = dist.get_rank(group)

    def _do_gather(tensor):
        if this_rank == dst:
            tensor_list = [torch.empty_like(tensor) for _ in range(world_size)]
        else:
            tensor_list = None
        dist.gather(tensor, tensor_list, dst=dst, group=group)
        return torch.cat(tensor_list, dim=cat_dim)

    if isinstance(container, dict):
        gathered = dict()
        for k, v in container.items():
            v = _do_gather(v)
            gathered[k] = v
        return gathered
    elif isinstance(container, (list, tuple)):
        gathered = [_do_gather(v) for v in container]
        if isinstance(container, tuple):
            gathered = tuple(gathered)
        return gathered
    else:
        # if not a dict, list, tuple, expect a singular tensor
        assert isinstance(container, torch.Tensor)
        return _do_gather(container)



import cv2
import timm
import torch
import numpy as np

from ... import core as mc
from .. import detector as md

# pylint: disable=missing-function-docstring
class PixelwiseClassifier(torch.nn.Module):
    """Basic segmentation model."""

    def __init__(self, n_features, n_classes):
        super().__init__()
        self.conv = torch.nn.Conv2d(
            in_channels=n_features, out_channels=n_classes, kernel_size=1, bias=True
        )
        self.act = torch.nn.Sigmoid()

    def forward(self, x):
        y = self.conv(x)
        y = self.act(y)
        return y


class Segmenter(torch.nn.Module):
    """A classifier using segmentation on bounding boxes."""

    def __init__(
        self,
        n_classes=5,
        downsample=4,
        model_name="efficientnet_lite0",
        pretrained=True,
        positive_agg=torch.max,
        negative_agg=torch.max,
    ):
        super().__init__()
        self.backbone = timm.create_model(
            model_name=model_name, pretrained=pretrained, features_only=True
        )
        self.downsample = downsample
        self.classifier = PixelwiseClassifier(
            n_features=sum(
                [f.shape[1] for f in self.backbone(torch.rand(1, 3, 128, 128))]
            ),
            n_classes=n_classes,
        )
        self.positive_agg = positive_agg
        self.negative_agg = negative_agg
        self.eval()

    def forward(self, x, target=None):
        y = self.backbone(x)
        y = [
            torch.nn.functional.interpolate(
                f, size=(x.shape[2] // self.downsample, x.shape[3] // self.downsample)
            )
            for f in y
        ]
        y = torch.cat(y, dim=1)
        y = self.classifier(y)
        if self.training:
            binary_inputs_list = []
            if any(ti["negative_pixels"] for ti in target):
                binary_inputs_list.append(
                    torch.cat(
                        mc.utils.flatten(
                            [
                                [
                                    torch.unsqueeze(
                                        self.negative_agg(
                                            torch.masked_select(yi[c], mask)
                                        ),
                                        0,
                                    )
                                    for mask, c in ti["negative_pixels"]
                                ]
                                for yi, ti in zip(y, target)
                            ]
                        )
                    )
                )
            if any(ti["positive_bboxes"] for ti in target):
                binary_inputs_list.append(
                    torch.cat(
                        mc.utils.flatten(
                            [
                                [
                                    torch.unsqueeze(
                                        self.positive_agg(yi[c, y1:y2, x1:x2]), 0  # type: ignore
                                    )
                                    for x1, y1, x2, y2, c in ti["positive_bboxes"]
                                ]
                                for yi, ti in zip(y, target)
                            ]
                        )
                    )
                )
            binary_inputs = torch.cat(binary_inputs_list)
            binary_target = torch.cat(
                [
                    torch.cat(
                        [
                            torch.zeros(len(ti["negative_pixels"]), device=x.device)
                            for ti in target
                        ]
                    ),
                    torch.cat(
                        [
                            torch.ones(len(ti["positive_bboxes"]), device=x.device)
                            for ti in target
                        ]
                    ),
                ]
            )
            return {
                "loss": torch.nn.functional.binary_cross_entropy(
                    input=binary_inputs, target=binary_target, reduction="mean"
                )
            }
        return y


def compute_boxes(scores, threshold):
    binary = (scores > threshold).astype("uint8")
    contours = cv2.findContours(
        binary, mode=cv2.RETR_EXTERNAL, method=cv2.CHAIN_APPROX_NONE
    )[0]
    x1y1x2y2 = np.array(
        [[c[:, 0].min(axis=0), c[:, 0].max(axis=0) + 1] for c in contours]
    ).reshape(-1, 4)
    return [
        (x1, y1, x2, y2, scores[y1:y2, x1:x2].mean()) for x1, y1, x2, y2 in x1y1x2y2
    ]


class AggregatedSegmentation(md.Detector):
    """A custom detector using segmentation to detect bounding boxes."""

    def __init__(
        self,
        annotation_config,
        device="cpu",
        resize_method: md.ResizeMethod = "fit",
        downsample: int = 4,
        model_name="efficientnet_lite0",
        pretrained_backbone: bool = True,
        positive_agg=torch.max,
        negative_agg=torch.max,
    ):
        self.resize_method = resize_method
        self.annotation_config = annotation_config
        self.model = Segmenter(
            n_classes=len(annotation_config),
            downsample=downsample,
            model_name=model_name,
            pretrained=pretrained_backbone,
            positive_agg=positive_agg,
            negative_agg=negative_agg,
        )
        self.set_device(device)
        self.set_input_shape(width=256, height=256)

    @property
    def training_model(self):
        """Training model for this detector."""
        return self.model

    def compute_inputs(self, images):
        images = np.float32(images) / 255.0
        images -= self.model.backbone.default_cfg["mean"]  # type: ignore
        images /= self.model.backbone.default_cfg["std"]  # type: ignore
        return (
            torch.tensor(images, dtype=torch.float32)
            .permute(0, 3, 1, 2)
            .to(self.device)
        )

    def invert_targets(self, y, threshold=0.5, **kwargs):
        annotations = []
        ds = self.model.downsample
        for yi in y.detach().cpu().numpy():
            current_annotations = []
            for labelIdx, category in enumerate(self.annotation_config):
                current_annotations.extend(
                    [
                        mc.Annotation(
                            category=category,
                            x1=x1 * ds,
                            y1=y1 * ds,
                            x2=x2 * ds,
                            y2=y2 * ds,
                            score=score,
                        )
                        for x1, y1, x2, y2, score in compute_boxes(
                            yi[labelIdx], threshold=threshold
                        )
                    ]
                )
            annotations.append(current_annotations)
        return annotations

    def set_input_shape(self, width, height):
        self._input_shape = (height, width, 3)

    def serve_module_string(self, enable_flexible_size=True):
        raise NotImplementedError

    @property
    def anchor_boxes(self):
        raise NotImplementedError

    @property
    def serve_module_index(self):
        raise NotImplementedError

    @property
    def input_shape(self):
        return self._input_shape

    def compute_targets(self, annotation_groups):
        height, width = self.input_shape[:2]
        ds = self.model.downsample
        height_ds, width_ds = height // ds, width // ds
        targets = []
        for g in annotation_groups:
            positive_bboxes = [
                np.floor(np.array([x1, y1, x2, y2]) / ds).astype("int32").tolist()
                + [cIdx]
                for x1, y1, x2, y2, cIdx in self.annotation_config.bboxes_from_group(g)
            ]
            # Make sure each box is at least one downsampled pixel wide.
            positive_bboxes = [
                (x1, y1, max(x2, x1 + 1), max(y2, y1 + 1), cIdx)
                for x1, y1, x2, y2, cIdx in positive_bboxes
            ]
            negative_pixels = np.ones(
                (len(self.annotation_config), height_ds, width_ds), dtype="bool"
            )
            for x1, y1, x2, y2, cIdx in positive_bboxes:
                negative_pixels[cIdx, y1:y2, x1:x2] = False
            targets.append(
                {
                    "positive_bboxes": positive_bboxes,
                    "negative_pixels": [
                        (torch.tensor(p).to(self.device), cIdx)
                        for cIdx, p in enumerate(negative_pixels)
                        if p.any()
                    ],
                }
            )
        return targets



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# Object detection data loaders and libraries are mostly based on RetinaNet:
# https://github.com/tensorflow/tpu/tree/master/models/official/retinanet
from .argmax_matcher import ArgMaxMatcher
from .box_coder import FasterRcnnBoxCoder
from .box_list import BoxList
from .matcher import Match
from .region_similarity_calculator import IouSimilarity
from .target_assigner import TargetAssigner
